﻿using System;
using System.Collections.Generic;
using System.Text;
using NGinnBPM.Lib.Data;
using NGinnBPM.Lib.Schema;
using NGinnBPM.Lib.Interfaces;
using NGinnBPM.Lib.Exceptions;
using NGinnBPM.Runtime.Events;
using NLog;
using System.Diagnostics;
using System.Collections;
using System.Runtime.Serialization;
using NGinnBPM.Runtime.Messages;
using NGinnBPM.MessageBus;

namespace NGinnBPM.Runtime
{

    /// <summary>
    /// Multi-instance task...
    /// Problem 1. How to enable tasks - sequentially, or in a single transaction?
    /// Problem 2. How to handle child task failure/cancellation
    /// </summary>
    [DataContract]
    public class MultiTaskInstance : TaskInstance,
        ITaskMessageHandler<TaskEnabled>,
        ITaskMessageHandler<TaskSelected>,
        ITaskMessageHandler<TaskCompleted>,
        ITaskMessageHandler<TaskCancelled>,
        ITaskMessageHandler<TaskFailed>,
        ITaskMessageHandler<EnableTaskTimeout>,
        ITaskMessageHandler<CancelTaskTimeout>
    {
        private List<TransitionInfo> _childTasks = new List<TransitionInfo>();

        /// <summary>
        /// If false, all child tasks must complete. If some child fails, whole multi task instance will fail also.
        /// If true, some of them can fail.
        /// </summary>
        [DataMember]
        [TaskParameter(IsInput=true, Required=false)]
        public bool AllowChildTaskFailures { get; set; }

        [DataMember(IsRequired=true)]
        public IList<TransitionInfo> ChildTransitions
        {
            get { return _childTasks; }
            set { _childTasks = new List<TransitionInfo>(value); }
        }

        public TransitionInfo GetTransitionInfo(string instanceId)
        {
            foreach (TransitionInfo ti in _childTasks)
            {
                if (ti.InstanceId == instanceId)
                    return ti;
            }
            return null;
        }

        public override void Enable(Dictionary<string, object> inputData)
        {
            List<Dictionary<string, object>> lst = new List<Dictionary<string, object>>();
            lst.Add(new Dictionary<string, object>(inputData));
            Enable(lst);
        }

       

        /// <summary>
        /// Do we have a race condition here?
        /// When enabling, the task is not yet persisted. So if it enables child tasks,
        /// we cannot accept messages from them until we are persisted as well...
        /// How to handle this???
        /// Current approach is to set a lock on instance ID so incoming child events will  have to wait
        /// until the lock is released...
        /// Problem: current task is not saved when enabling child tasks, however we are giving them parent
        /// id. Their parent is not in database, so they cannot access parent task instance...
        /// Solution - do it in async way........ not very performant, however. At least unless we use
        /// better messaging
        /// </summary>
        /// <param name="inputData"></param>
        public void Enable(ICollection<Dictionary<string, object>> inputData)
        {
            RequireActivation(true);
            //TODO: validate the data
            if (inputData.Count == 0)
                throw new TaskDataInvalidException(TaskId, InstanceId, "No input data for multi-instance task");

            foreach (Dictionary<string, object> dob in inputData)
            {
                TransitionInfo ti = new TransitionInfo();
                ti.Status = TransitionStatus.Enabling;
                ti.InstanceId = AllocateNewTaskInstanceId(TaskId);
                _childTasks.Add(ti);
                
                EnableTaskMessage etm = new EnableTaskMessage();
                etm.InputData = dob;
                etm.NewTaskInstanceId = ti.InstanceId;
                etm.CorrelationId = ti.InstanceId;
                etm.ParentTaskInstanceId = this.InstanceId;
                etm.ProcessDefinition = this.ProcessDefinitionId;
                etm.ProcessInstanceId = this.ProcessInstanceId;
                etm.TaskId = this.TaskId;
                Context.MessageBus.NewMessage(etm).Publish();
                Context.MessageBus.NewMessage(new EnableTaskTimeout { TargetTaskInstanceId = this.InstanceId })
                    .SetCorrelationId(ti.InstanceId)
                    .SetDeliveryDate(DateTime.Now.AddDays(3))
                    .Publish();
            }
            Status = TaskStatus.Enabled;
            EnabledDate = DateTime.Now;
        }

        protected void EnableInternal(ICollection<Dictionary<string, object>> inputData)
        {
            try
            {
                RequireActivation(true);
                if (Status != TaskStatus.Enabled && Status != TaskStatus.Selected)
                {
                    log.Info("Will not enable child task because parent is inactive");
                    return;
                }
                if (inputData.Count == 0)
                    throw new TaskDataInvalidException(TaskId, InstanceId, "No input data for multi-instance task");
                log.Info("Will create {0} instances of task {1}", inputData.Count, TaskId);
                try
                {
                    int cnt = 0;
                    foreach (Dictionary<string, object> dob in inputData)
                    {
                        TransitionInfo ti = new TransitionInfo();
                        ti.Status = TransitionStatus.Enabled;
                        ti.TaskId = TaskId;
                        _childTasks.Add(ti);
                        ti.InstanceId = Context.Environment.EnableTask(ProcessDefinitionId, TaskId, dob, this.InstanceId, this.ProcessInstanceId);
                        log.Info("Created new instance: {0}", ti.InstanceId);
                    }
                }
                catch (Exception ex)
                {
                    log.Error("Error creating multi-instance tasks: {0}", ex);
                    //todo: compensate here???
                    foreach (TransitionInfo ti in _childTasks)
                    {
                        if (ti.InstanceId != null && ti.IsTransitionActive)
                        {
                            try
                            {
                                log.Info("Cancelling already created multi-task instance: {0}", ti.InstanceId);
                                Context.Environment.CancelTaskInstance(ti.InstanceId);
                            }
                            catch (Exception ex2)
                            {
                                log.Error("Failed to cancel multi-task instance {0}: {1}", ti.InstanceId, ex2);
                            }
                        }
                    }
                    throw;
                }
                Status = TaskStatus.Enabled;
            }
            catch (Exception ex)
            {
                log.Error("Multi-instance failed to start child tasks: {0}", ex);
                DefaultHandleTaskFailure(ex.Message, false);
            }
        }

       
        protected TransitionInfo GetTransition(string instanceId)
        {
            foreach (TransitionInfo ti in _childTasks)
                if (ti.InstanceId == instanceId) return ti;
            return null;
        }

        

        protected void OnTransitionStatusChanged(string transitionId)
        {
            if (IsAnyTransitionActiveYet())
            {
                return;
            }
            log.Info("There are no active transitions. Completing task.");
            if (Status == TaskStatus.Cancelling)
            {
                DefaultHandleTaskCancelled();
            }
            else
            {
                Debug.Assert(Status == TaskStatus.Enabled || Status == TaskStatus.Selected);
                List<Dictionary<string, object>> lst = new List<Dictionary<string, object>>();
                foreach (TransitionInfo ti in _childTasks)
                {
                    if (ti.Status == TransitionStatus.Completed)
                    {
                        Dictionary<string, object> dob = ti.OutputData;
                        if (dob == null) dob = new Dictionary<string, object>();
                        lst.Add(ti.OutputData);
                    }
                    else
                    {
                        log.Info("Skipping failed or cancelled child task: {0}", ti);
                    }
                }
                Dictionary<string, object> dob2 = new Dictionary<string, object>();
                dob2["multiInstanceResults"] = lst;
                DefaultHandleTaskFinished(dob2);
            }
        }

        protected override void DefaultHandleTaskFinished(Dictionary<string, object> outData)
        {
            throw new NotImplementedException("To nie działa");
            Task tsk = ParentProcess.GetTask(TaskId);
            StructDef sd = tsk.GetOutputDataSchema();
            DataUtil.Validate(outData, sd);
            FinishedDate = DateTime.Now;
            Status = TaskStatus.Completed;
            TaskCompleted tce = new TaskCompleted();
            tce.SourceTaskInstanceId = this.InstanceId;
            tce.ParentTaskInstanceId = this.ParentTaskInstanceId;
            tce.OutputData = outData;
            Context.MessageBus.Notify(tce);
        }

        private void CancelRemainingActiveChildTasks()
        {
            foreach (TransitionInfo ti in ChildTransitions)
            {
                if (ti.Status == TransitionStatus.Enabled ||
                    ti.Status == TransitionStatus.Enabling ||
                    ti.Status == TransitionStatus.Started)
                {
                    CancelTaskMessage ctm = new CancelTaskMessage();
                    ctm.TargetTaskInstanceId = ti.InstanceId;
                    Context.MessageBus.Notify(ctm);
                    ti.Status = TransitionStatus.Cancelling;
                }
                else if (ti.Status == TransitionStatus.FailedActive)
                {
                    ti.Status = TransitionStatus.Cancelled;
                }
                Debug.Assert(ti.Status == TransitionStatus.Completed || ti.Status == TransitionStatus.Failed || ti.Status == TransitionStatus.Cancelled || ti.Status == TransitionStatus.Cancelling);
            }
        }

        public override void Cancel()
        {
            lock(this)
            {
                if (Status == TaskStatus.Cancelled || Status == TaskStatus.Cancelling)
                    return;
                else if (Status != TaskStatus.Enabled && Status != TaskStatus.Selected) throw new InvalidTaskStatusException(InstanceId, "Invalid status");
                CancelRemainingActiveChildTasks();
                if (!IsAnyTransitionActiveYet())
                {
                    base.Cancel();
                    Debug.Assert(Status == TaskStatus.Cancelled);
                }
                else
                {
                    Status = TaskStatus.Cancelling;
                    Context.MessageBus.NewMessage(new CancelTaskTimeout { TargetTaskInstanceId = InstanceId })
                        .SetDeliveryDate(DateTime.Now.AddHours(24))
                        .Publish();
                }
            }
        }

        private bool IsAnyTransitionActiveYet()
        {
            foreach (TransitionInfo ti in _childTasks)
            {
                if (ti.IsTransitionActive) return true;
            }
            return false;
        }

        public override void Fail(string errorInformation)
        {
            CancelRemainingActiveChildTasks();
            base.Fail(errorInformation);
        }

        public override void Complete(string finishedBy, Dictionary<string, object> updatedData)
        {
            CancelRemainingActiveChildTasks();
            base.Complete(finishedBy, updatedData);
        }

        /// <summary>
        /// creates a new task instance id for a child task.
        /// The id generated should be unique.
        /// This way composite tasks can control what ids are assigned to child tasks.
        /// </summary>
        /// <param name="taskId"></param>
        /// <returns></returns>
        protected string AllocateNewTaskInstanceId(string taskId)
        {
            int n;
            lock (this)
            {
                n = ++ChildCounter;
            }
            string s = string.Format("{0}.{1}", InstanceId, n);
            if (this.GetTransitionInfo(s) != null) throw new Exception("Child task instance ID is duplicate. Oh my!");
            return s;
        }

        [DataMember]
        public int ChildCounter { get; set; }

        #region IMessageConsumer<TaskEnabledMessage> Members

        public void Handle(TaskEnabled message)
        {
            RequireActivation(true);
            TransitionInfo ti = GetTransition(message.CorrelationId);
            if (ti == null) throw new TaskRuntimeException("Child transition not found: " + message.CorrelationId).SetInstanceId(InstanceId);
            if (ti.Status == TransitionStatus.Enabling)
            {
                ti.Status = TransitionStatus.Enabled;
                if (ti.InstanceId != message.SourceTaskInstanceId)
                {
                    log.Info("Child task instance id changed {0}->{1}", ti.InstanceId, message.SourceTaskInstanceId);
                    ti.InstanceId = message.SourceTaskInstanceId;
                }
            }
            else
            {
                log.Debug("Ignoring message: {0}", message);
            }
        }

        #endregion

        

        #region IMessageConsumer<TaskSelected> Members

        public void Handle(TaskSelected message)
        {
            RequireActivation(true);
            if (message.ParentTaskInstanceId != this.InstanceId) throw new Exception();
            TransitionInfo ti = GetTransition(message.SourceTaskInstanceId);
            if (ti == null) throw new Exception();
            lock (this)
            {
                log.Info("Child task {0} has been started", ti.InstanceId);
                if (ti.Status == TransitionStatus.Started)
                    return;
                if (ti.Status != TransitionStatus.Enabled && ti.Status != TransitionStatus.Enabling)
                {
                    log.Warn("Transition {0} ({1}) is not enabled: {2}", ti.InstanceId, ti.TaskId, ti.Status);
                    return;
                }
                ti.Status = TransitionStatus.Started;
                OnTransitionStatusChanged(ti.InstanceId);
            }
        }

        #endregion

        #region IMessageConsumer<TaskCompleted> Members

        public void Handle(TaskCompleted message)
        {
            RequireActivation(true);
            if (message.ParentTaskInstanceId != this.InstanceId) throw new Exception();
            TransitionInfo ti = GetTransition(message.SourceTaskInstanceId);
            if (ti == null) throw new Exception();
            lock (this)
            {
                log.Info("Child task {0} has been completed", ti.InstanceId);
                if (ti.Status == TransitionStatus.Completed)
                    return;
                if (!ti.IsTransitionActive)
                {
                    log.Warn("Transition {0} ({1}) is not active: {2}", ti.InstanceId, ti.TaskId, ti.Status);
                    return;
                }
                ti.Status = TransitionStatus.Completed;
                ti.OutputData = message.OutputData;
                if (ti.OutputData == null)
                {
                    log.Info("No output data returned from child task {0}", ti.InstanceId);
                    ti.OutputData = new Dictionary<string, object>();
                }
                OnTransitionStatusChanged(ti.InstanceId);
                return;
            }
        }

        #endregion

        #region IMessageConsumer<TaskCancelled> Members

        public void Handle(TaskCancelled message)
        {
            RequireActivation(true);
            if (message.ParentTaskInstanceId != this.InstanceId) throw new Exception();
            TransitionInfo ti = GetTransition(message.SourceTaskInstanceId);
            if (ti == null) throw new Exception();
            lock (this)
            {
                log.Info("Child task {0} has been cancelled", ti.InstanceId);
                if (ti.Status == TransitionStatus.Cancelled)
                    return;
                if (!ti.IsTransitionActive)
                {
                    log.Warn("Transition {0} ({1}) is not active: {2}", ti.InstanceId, ti.TaskId, ti.Status);
                    return;
                }
                ti.Status = TransitionStatus.Cancelled;
                OnTransitionStatusChanged(ti.InstanceId);
                return;
            }
        }

        #endregion

        #region IMessageConsumer<TaskFailed> Members

        public void Handle(TaskFailed message)
        {
            RequireActivation(true);
            if (message.ParentTaskInstanceId != this.InstanceId) throw new Exception();
            TransitionInfo ti = GetTransition(message.SourceTaskInstanceId);
            if (ti == null) throw new Exception();
            lock (this)
            {
                log.Info("Child task {0} has failed: {1}", ti.InstanceId, message.ErrorInfo);
                if (ti.Status == TransitionStatus.Failed)
                    return;
                if (!ti.IsTransitionActive)
                {
                    log.Warn("Transition {0} ({1}) is not active: {2}", ti.InstanceId, ti.TaskId, ti.Status);
                    return;
                }
                if (message.IsExpected)
                {
                    ti.Status = TransitionStatus.Failed;
                    if (!AllowChildTaskFailures)
                    {
                        Fail(message.ErrorInfo);
                        return;
                    }
                }
                else
                {
                    ti.Status = TransitionStatus.FailedActive;
                }
                OnTransitionStatusChanged(ti.InstanceId);
                return;
            }
        }

        #endregion

        #region IMessageConsumer<EnableTaskTimeout> Members

        public void Handle(EnableTaskTimeout message)
        {
            RequireActivation(true);
            string corrid = Context.MessageBus.CurrentMessageInfo.CorrelationId;
            TransitionInfo ti = GetTransition(corrid);
            if (ti == null) throw new TaskRuntimeException("Child transition not found: " + corrid).SetInstanceId(InstanceId);
            if (ti.Status == TransitionStatus.Enabling)
            {
                throw new NotImplementedException();
                Fail(string.Format("Child task enable timed out: {0}", ti.InstanceId));
                //ti.Status = TransitionStatus.Enabled;
            }
            else
            {
                log.Debug("Ignoring message: {0}", message);
            }
        }

        #endregion

        #region IMessageConsumer<CancelTaskTimeout> Members

        public void Handle(CancelTaskTimeout message)
        {
            RequireActivation(true);
            if (Status != TaskStatus.Cancelling)
            {
                log.Info("Ignoring cancel timeout.");
                return;
            }
            if (IsAnyTransitionActiveYet())
            {
                log.Warn("Task {0} cancellation timeout. Ignoring child transitions that haven't yet cancelled and forcing the cancellation", InstanceId);
                DefaultHandleTaskCancelled();
                Debug.Assert(Status == TaskStatus.Cancelled);
            }
            else
            {
                //we should not be cancelling anymore - a bug!!!
                log.Error("Task {0}: cancellation timeout, but there are no active transitions. Task should have cancelled");
                Debug.Assert(false);
                DefaultHandleTaskCancelled();
            }
        }

        #endregion
    }
}
